use anyhow::Context;
use quickpeep_index::backend::tantivy::TantivyBackend;
use quickpeep_index::backend::Backend;
use quickpeep_index::config::BackendConfig;
use serde::{Deserialize, Serialize};
use std::path::{Path, PathBuf};

#[derive(Serialize, Deserialize, Debug, Clone)]
/// Config for an indexer. All paths are relative to the config file if needed, but will be resolved
/// when loading.
pub struct IndexerConfig {
    /// Path to seeds
    pub seed_dir: PathBuf,

    /// Path to the icon store
    pub icon_store: PathBuf,

    /// Configuration about which backend to use.
    pub backend: BackendConfig,
}

impl IndexerConfig {
    /// Loads a config at the specified path.
    /// Will resolve all the paths in the IndexerConfig for you.
    pub fn load(path: &Path) -> anyhow::Result<IndexerConfig> {
        let config_dir = path.parent().context("Can't get parent of config file.")?;
        let bytes = std::fs::read(path)?;
        let mut indexer_config: IndexerConfig = toml::from_slice(&bytes)?;

        indexer_config.seed_dir = config_dir.join(indexer_config.seed_dir);
        indexer_config.icon_store = config_dir.join(indexer_config.icon_store);
        match &mut indexer_config.backend {
            BackendConfig::Tantivy(tantivy) => {
                tantivy.index_dir = config_dir.join(&tantivy.index_dir);
            }
            BackendConfig::Meili(_) => {}
        }

        Ok(indexer_config)
    }

    pub fn open_indexer_backend(&self) -> anyhow::Result<Box<dyn Backend>> {
        match &self.backend {
            BackendConfig::Tantivy(tantivy) => {
                Ok(Box::new(TantivyBackend::open(&tantivy.index_dir)?))
            }
            BackendConfig::Meili(_) => {
                todo!()
            }
        }
    }
}
