use crate::storage::mdbx_helper_types::{MdbxBare, MdbxString};
use crate::storage::records::{AllowedDomainRecord, OnHoldUrlRecord, WeedDomainRecord};
use crate::storage::RakerTxn;
use anyhow::Context;
use libmdbx::{Database, WriteFlags, RW};
use log::warn;
use reqwest::Url;

/// Runs one big transaction that:
/// - scans on-hold URLs
///     - moves 'allowed' ones to the queue
///     - deletes 'weeds'
///     - leaves unknown ones alone
///
/// Ideally should be applied after importing seeds and weeds on an existing database.
pub fn reapply_seeds_and_weeds_to_on_hold_urls(txn: RakerTxn<RW>) -> anyhow::Result<()> {
    struct DomainState {
        pub domain: String,
        pub allowed_domain_record: Option<AllowedDomainRecord>,
        pub weed_domain_record: Option<WeedDomainRecord>,
    }

    let urls_on_hold: &Database = &txn.mdbx.borrow_dbs().urls_on_hold;

    let mut domain_state = None;

    // Scan through the on-hold URLs
    let mut cur = txn.mdbx_txn.cursor(urls_on_hold)?;
    let mut first_iteration = true;

    while let Some((MdbxString(domain_then_url), MdbxBare(record))) = if first_iteration {
        first_iteration = false;
        cur.first::<MdbxString, MdbxBare<OnHoldUrlRecord>>()
    } else {
        cur.next::<MdbxString, MdbxBare<OnHoldUrlRecord>>()
    }? {
        let mut split = domain_then_url.as_ref().split("\n");
        let domain = split.next().context("No first split..?")?;
        let url_str = split.next().context("No URL")?;

        // Is the domain new?
        if domain_state
            .as_ref()
            .map(|ds: &DomainState| &ds.domain != domain)
            .unwrap_or(true)
        {
            // Then load the relevant records for it.
            domain_state = Some(DomainState {
                domain: domain.to_owned(),
                allowed_domain_record: txn.get_allowed_domain_record(domain)?,
                weed_domain_record: txn.get_weed_domain_record(domain)?,
            });
        }

        let url = Url::parse(url_str)?;

        let domain_state = domain_state.as_ref().unwrap();
        let is_allowed = domain_state
            .allowed_domain_record
            .as_ref()
            .map(|adr: &AllowedDomainRecord| adr.applies_to_url(&url))
            .unwrap_or(false);
        let is_weed = domain_state
            .weed_domain_record
            .as_ref()
            .map(|wdr: &WeedDomainRecord| wdr.applies_to_url(&url))
            .unwrap_or(false);

        match (is_allowed, is_weed) {
            (false, false) => { /* nop */ }
            (true, true) => {
                warn!(
                    "Ambiguous: {:?} is both mentioned by a seed and a weed. Ignoring.",
                    url
                );
            }
            (true, false) => {
                // ALLOWED
                // Make it a queued URL
                txn.enqueue_url(url_str, None, record.queue_record.intent)?;
                cur.del(WriteFlags::empty())?;
            }
            (false, true) => {
                // WEED
                // Just delete
                cur.del(WriteFlags::empty())?;
            }
        }
    }

    txn.commit()?;
    Ok(())
}
